#!/usr/bin/env python
import cmath
import importlib
import simple_pid

import cv2
import rospy
from sensor_msgs.msg import Joy, Image, JointState
from cv_bridge import CvBridge

COMMAND_MODULE_PATH = "/home/k-lab/catkin_ws/devel/lib/python3/dist-packages/rtc2021/msg/_Command.py"
spec = importlib.util.spec_from_file_location("Command", COMMAND_MODULE_PATH)
foo = importlib.util.module_from_spec(spec)
spec.loader.exec_module(foo)
Command = foo.Command


class RobotState:
    auto = False
    arm_rotate = 0
    arm_0 = 0
    arm_1 = 0
    arm_fist = 0


robot_state = RobotState()


class ImageState:
    image = 0


image_state = ImageState()


class GamepadState:
    LS = [0, 0]
    RS = [0, 0]
    LT = 0
    RT = 0
    DPAD = [0, 0]
    A = 0
    B = 0
    X = 0
    Y = 0
    LB = 0
    RB = 0
    Window = 0
    Menu = 0


gamepad_state = GamepadState()


def callback_gamepad(data):
    gamepad_state.LS = data.axes[0:2]
    gamepad_state.RS = data.axes[2:4]
    gamepad_state.LT = data.axes[4]
    gamepad_state.RT = data.axes[5]
    gamepad_state.DPAD = data.axes[6:8]
    gamepad_state.A = data.buttons[0]
    gamepad_state.B = data.buttons[1]
    gamepad_state.X = data.buttons[2]
    gamepad_state.Y = data.buttons[3]
    gamepad_state.LB = data.buttons[4]
    gamepad_state.RB = data.buttons[5]
    gamepad_state.Window = data.buttons[6]
    gamepad_state.Menu = data.buttons[7]


def callback_camera(data):
    image_state.image = data


def control_manual(robot: RobotState, gamepad: GamepadState, image: ImageState, command: Command) -> Command:
    if gamepad.Window:
        robot.auto = True
        return robot, gamepad, image, command

    x = gamepad.LS[0]
    y = gamepad.LS[1]

    gamma = 2
    theta = 2

    WL = y + x * (theta / (abs(y) + gamma))
    WR = y - x * (theta / (abs(y) + gamma))

    WL = WL if abs(WL) <= 1 else WL / abs(WL)
    WR = WR if abs(WR) <= 1 else WR / abs(WR)

    command.wheel_left = WL
    command.wheel_right = WR

    robot.arm_rotate -= gamepad.DPAD[0] / 100 if (abs(robot.arm_rotate - gamepad.DPAD[0]/100) < 0.95) else 0

    robot.arm_0 -= gamepad.DPAD[1] / 100 if (abs(robot.arm_0 - gamepad.DPAD[1]/100) < 0.95) else 0

    robot.arm_1 -= gamepad.RS[1] / 100 if (abs(robot.arm_1 - gamepad.RS[1]/100) < 0.95) else 0

    if gamepad.X:
        robot.arm_rotate = 0
        robot.arm_0 = 0
        robot.arm_1 = 0
        robot.arm_fist = 0

    command.arm_rotate = robot.arm_rotate
    command.arm_0 = robot.arm_0
    command.arm_1 = robot.arm_1

    return robot, gamepad, image, command


pid = simple_pid.PID(1, 0.1, 0.05)
pid.output_limits = (-1, 1)


def control_auto(robot: RobotState, gamepad: GamepadState, image: ImageState, command: Command) -> Command:
    if gamepad.Menu:
        robot.auto = False
        return robot, gamepad, image, command

    if image_state.image:
        br = CvBridge()
        cv_image = br.imgmsg_to_cv2(image_state.image)
        cv_image = cv_image[80:480, 0:640]
        gray = cv2.cvtColor(cv_image, cv2.COLOR_BGR2GRAY)
        blur = cv2.GaussianBlur(gray, (5, 5), 0)
        ret, thresh = cv2.threshold(blur, 80, 255, cv2.THRESH_BINARY_INV)
        # thresh = cv2.erode(thresh, None, iterations=3)
        # thresh = cv2.dilate(thresh, None, iterations=8)
        contours, hierarchy = cv2.findContours(thresh.copy(), 1, cv2.CHAIN_APPROX_NONE)

        control_signal = 0.0
        WL = 0
        WR = 0

        if len(contours):
            c = max(contours, key=cv2.contourArea)
            m = cv2.moments(c)

            cx = int(m['m10'] / (m['m00'] + 1))
            cy = int(m['m01'] / (m['m00'] + 1))
            cv2.line(cv_image, (cx, 0), (cx, 640), (255, 0, 0), 1)
            cv2.line(cv_image, (0, cy), (480, cy), (255, 0, 0), 1)
            cv2.drawContours(thresh, contours, -1, (0, 255, 0), 1)

            delta = (cx - 320) / 320

            cv2.imshow("lol", thresh)
            cv2.waitKey(20)

            control_signal = pid(delta)

            m = (5000 + cv2.contourArea(max(contours, key=cv2.contourArea))) / 25000

            WL = (0.5 - control_signal) * m
            WR = (0.5 + control_signal) * m
        else:
            WL = -0.25
            WR = -0.25

        WR = WR if abs(WR) <= 1 else abs(WR) / WR
        WL = WL if abs(WL) <= 1 else abs(WL) / WL

        command.wheel_left = WL
        command.wheel_right = WR
    return robot, gamepad, image, command


def node_control():
    global robot_state, gamepad_state, image_state
    rospy.init_node("node_control")
    rospy.Subscriber("/usb_cam/image_raw", Image, callback=callback_camera)
    rospy.Subscriber("joy", Joy, callback=callback_gamepad)
    pub = rospy.Publisher("/node_tx/command", Command, queue_size=1)
    joint = rospy.Publisher("/node_control/joint", JointState, queue_size=1)
    rate = rospy.Rate(100)
    while not rospy.is_shutdown():
        command = Command()
        if robot_state.auto:
            robot, gamepad, image, command = control_auto(robot_state, gamepad_state, image_state, command)
        else:
            robot, gamepad, image, command = control_manual(robot_state, gamepad_state, image_state, command)
        robot_state = robot
        gamepad_state = gamepad
        image_state = image

        pub.publish(command)

        hello_str = JointState()
        hello_str.header.stamp = rospy.Time.now()
        hello_str.name = ['base_link_to_virtual_arm_rotate', 'virtual_arm_rotate_to_virtual_arm_0_motor', 'virtual_arm_0_to_virtual_arm_1_motor']
        hello_str.position = [robot.arm_rotate * 140 / 360 * 3.14, -robot.arm_0 * 3.14 / 2, -robot.arm_1 * 3.14 / 2]
        hello_str.velocity = []
        hello_str.effort = []
        joint.publish(hello_str)

        rate.sleep()


if __name__ == "__main__":
    try:
        node_control()
    except rospy.ROSInterruptException:
        pass
