#!/usr/bin/env python3
from ctypes import Structure, c_int8
from typing import Optional, Tuple
import crc8
from serial import Serial
import rospy
import importlib.util
from sensor_msgs.msg import JointState


TELEMETRY_MODULE_PATH = "/home/k-lab/catkin_ws/devel/lib/python3/dist-packages/rtc2021/msg/_Telemetry.py"
spec = importlib.util.spec_from_file_location("Telemetry", TELEMETRY_MODULE_PATH)
foo = importlib.util.module_from_spec(spec)
spec.loader.exec_module(foo)
Telemetry = foo.Telemetry


class TelemetryConfig:
    PAYLOAD_SIZE = 12
    PACKAGE_PARSED_SIZE = 3*(PAYLOAD_SIZE + 4)//2
    RX_PATH = '/dev/serial/by-id/usb-FTDI_FT230X_Basic_UART_D308ZXNS-if00-port0'


class TelemetryPackage(Structure):
    _pack_ = 0
    _fields_ = [
        ("wheel_left", c_int8),
        ("wheel_right", c_int8),
        ("arm_rotate", c_int8),
        ("arm_0", c_int8),
        ("arm_1", c_int8),
        ("arm_fist", c_int8),
        ("led", c_int8),
        ("spike_left", c_int8),
        ("spike_right", c_int8),
        ("acc_x", c_int8),
        ("acc_y", c_int8),
        ("acc_z", c_int8)
    ]


def compute_check_sum(input_bytes: bytes) -> int:
    crc = crc8.crc8()
    crc.update(input_bytes)
    check_sum = int(str(crc.hexdigest()), 16)
    return check_sum


# Returns (True, valid_package) or (False, empty_bytearray)
def parse_package(input_bytes: Optional[bytes], payload_bytes_len: int) -> Tuple[bool, bytes]:
    # check len(input_bytes) < len(start_bytes) + len(payload_bytes) + len(check_sum)
    input_bytes_len = len(input_bytes)
    if input_bytes_len < payload_bytes_len + 4:
        return False, bytearray()

    # finite state machine for start bytes
    found_state = False
    for idx in range(input_bytes_len - payload_bytes_len - 3):
        if input_bytes[idx] == 0xaa:
            if found_state:
                payload_bytes = input_bytes[idx + 1:idx + payload_bytes_len + 2]
                actual_check_sum = input_bytes[idx + payload_bytes_len + 2]
                package_check_sum = compute_check_sum(payload_bytes)
                return package_check_sum == actual_check_sum, payload_bytes[1:]
            else:
                found_state = True
        else:
            found_state = False
    return False, bytearray()


def node_telemetry():
    while not rospy.is_shutdown():
        rospy.init_node("node_telemetry")
        rospy.loginfo("Telemetry started")

        pub = rospy.Publisher(
            name="/node_telemetry/telemetry",
            data_class=Telemetry,
            queue_size=10
        )
        joint = rospy.Publisher("/node_control/joint", JointState, queue_size=1)
        try:
            ser = Serial(TelemetryConfig.RX_PATH, 57600)
            telemetry = Telemetry()
            while not rospy.is_shutdown():
                fetched_bytearray = ser.read(TelemetryConfig.PACKAGE_PARSED_SIZE)
                if len(fetched_bytearray) < TelemetryConfig.PACKAGE_PARSED_SIZE:
                    telemetry.exception = True
                    telemetry.header.stamp = rospy.get_rostime()
                    pub.publish(telemetry)
                    continue
                is_valid, package = parse_package(fetched_bytearray, TelemetryConfig.PAYLOAD_SIZE)
                if is_valid:
                    telemetry_package = TelemetryPackage.from_buffer_copy(package)
                    telemetry.wheel_left = telemetry_package.wheel_left / 127
                    telemetry.wheel_right = telemetry_package.wheel_right / 127
                    telemetry.arm_rotate = telemetry_package.arm_rotate / 127
                    telemetry.arm_0 = telemetry_package.arm_0 / 127
                    telemetry.arm_1 = telemetry_package.arm_1 / 127
                    telemetry.arm_fist = telemetry_package.arm_fist / 127
                    telemetry.led = telemetry_package.led / 255
                    telemetry.spike_left = telemetry_package.spike_left / 127
                    telemetry.spike_right = telemetry_package.spike_right / 127
                    telemetry.acc_x = telemetry_package.acc_x / 127
                    telemetry.acc_y = telemetry_package.acc_y / 127
                    telemetry.acc_z = telemetry_package.acc_z / 127
                    telemetry.exception = False
                    telemetry.header.stamp = rospy.get_rostime()
                    pub.publish(telemetry)
                    hello_str = JointState()
                    hello_str.header.stamp = rospy.Time.now()
                    hello_str.name = ['base_link_to_arm_rotate', 'arm_rotate_to_arm_0_motor', 'arm_0_to_arm_1_motor']
                    hello_str.position = [telemetry.arm_rotate*140/360*3.14, -telemetry.arm_0*3.14/2, -telemetry.arm_1*3.14/2]
                    hello_str.velocity = []
                    hello_str.effort = []
                    joint.publish(hello_str)

        except Exception as e:
            telemetry = Telemetry()
            telemetry.exception = True
            telemetry.header.stamp = rospy.get_rostime()
            pub.publish(telemetry)
            rospy.logerr(e)


if __name__ == "__main__":
    try:
        node_telemetry()
    except rospy.ROSInterruptException as e:
        pass
